from cosmosis.runtime import FunctionModule
from cosmosis.datablock import DataBlock
from cosmosis.gaussian_likelihood import GaussianLikelihood, SingleValueGaussianLikelihood
import numpy as np
import os

def test_gaussian():
    class MyLikelihood(GaussianLikelihood):
        x_section = "aaa"
        x_name = "a"
        y_section = "bbb"
        y_name = "b"
        like_name = "lll"

        def build_data(self):
            x_obs = np.array([1.0, 2.0, 3.0])
            y_obs = x_obs * 2
            return x_obs, y_obs

        def build_covariance(self):
            covmat = np.diag([0.1, 0.1, 0.1])
            return covmat

    mod = MyLikelihood.as_module("my")

    # no extra config info
    mod.setup({"my":{"include_norm":True}})

    block = DataBlock()
    block["aaa", "a"] = np.arange(5.)
    block["bbb", "b"] = np.arange(5.) * 2
    status = mod.execute(block)

    assert status == 0
    assert np.isclose(block["data_vector", "lll_chi2"], 0)
    assert np.isclose(block["data_vector", "lll_log_det"], 3*np.log(0.1))
    assert block["data_vector", "lll_n"] == 3

    assert np.isclose(block["likelihoods", "lll_like"], -3*np.log(0.1)/2)

def test_single_gaussian():

    class MySingleLikelihood(SingleValueGaussianLikelihood):
        section = "sec"
        name = "name"
        like_name = "xxx"
        mean = 3.0
        sigma = 0.1


    mod = MySingleLikelihood.as_module("my2")

    # no extra config info
    mod.setup({"my2":{"include_norm":True, "likelihood_only": False}})

    block = DataBlock()

    block["sec", "name"] = 4.0
    status = mod.execute(block)

    # check cholesky correctly calculated
    assert np.isclose(mod.data.chol, 0.1)

    assert status == 0
    assert np.isclose(block["data_vector", "xxx_chi2"], 100)
    assert np.isclose(block["data_vector", "xxx_log_det"], 2*np.log(0.1))
    assert block["data_vector", "xxx_n"] == 1

    assert np.isclose(block["data_vector", "xxx_theory"], 4.0)
    assert np.isclose(block["data_vector", "xxx_data"], 3.0)
    assert np.isclose(block["data_vector", "xxx_inverse_covariance"], 100.0)
    # sim should be within 10 sigma!
    assert 3.0 < block["data_vector", "xxx_simulation"] < 5.0

    assert np.isclose(block["likelihoods", "xxx_like"], -50.0 - np.log(0.1))


if __name__ == '__main__':
    test_gaussian()





import matplotlib
matplotlib.use('Agg')
import pylab
import configparser
import collections
from .utils import NoSuchParameter, section_code

class Plotter(object):
    colors=['blue','red','green','cyan','gray']
    def __init__(self, chain_data,latex_file=None, filetype="png", root_dir='.',prefix='',blind=False, fatal=False, **options):
        self._chain_data = chain_data
        all_names = set()
        for chain_datum in list(self._chain_data.values()):
            for name in list(chain_datum.keys()):
                all_names.add(name)
        self.all_names = sorted(list(all_names))
        self.load_latex(latex_file)
        self.nfile = len(chain_data)
        self.filetype=filetype
        self.options=options
        self.root_dir = root_dir
        self.prefix = prefix
        self.blind = blind
        self.fatal=fatal
        if self.prefix and not self.prefix.endswith('_'):
            self.prefix = self.prefix + "_"


    def command(self, command, *args, **kwargs):
        cmd = command.format(*args, **kwargs) + '\n'
        self._output_file.write(cmd)

    def blind_axes(self):
        pylab.tick_params(
            axis='both',          # changes apply to the x-axis
            which='both',      # both major and minor ticks are affected
            bottom='off',      # ticks along the bottom edge are off
            top='off',         # ticks along the top edge are off
            left='off',      # ticks along the bottom edge are off
            right='off',         # ticks along the top edge are off
            labelbottom='off', # labels along the bottom edge are off
            labeltop='off', # labels along the bottom edge are off
            labelleft='off', # labels along the bottom edge are off
            labelright='off') # labels along the bottom edge are off


    def load_latex(self, latex_file):
        self._display_names = {}
        latex_names = {}
        if latex_file is not None:
            latex_names = configparser.ConfigParser(strict=False)
            latex_names.read(latex_file)
        for i,col_name in enumerate(self.all_names):
            display_name=col_name
            if '--' in col_name:
                section,name = col_name.lower().split('--')
                try:
                    display_name = latex_names.get(section,name)
                except configparser.NoSectionError:
                    section = section_code(section)
                except configparser.NoOptionError:
                    pass                    
                try:
                    display_name = latex_names.get(section,name)
                except:
                    pass
            else:
                if col_name in ["LIKE","likelihood"]:
                    display_name=r"{\cal L}"
            self._display_names[col_name]=display_name

    def cols_for_name(self, name):
        cols = collections.OrderedDict()
        for filename, chain_datum in list(self._chain_data.items()):
            if name in list(chain_datum.keys()):
                cols[filename] = chain_datum[name]
        if not cols:
            raise NoSuchParameter(name)
        return cols

    def parameter_range(self, name):
        cols = self.cols_for_name(name)
        xmin = 1e30
        xmax = -1e30
        for col in list(cols.values()):
            if col.min() < xmin: xmin=col.min()
            if col.max() > xmax: xmax=col.max()
        if xmin==1e30 or xmax==-1e30:
            raise ValueError("Could not find col max/min - NaNs in chain?")
        return xmin,xmax

    def plot_1d_params(self, names):
        if not names:
            names = self.all_names
        for name in names:
            if name.lower()=='like' or name.lower()=='like': continue
            print("Plotting 1D curve for ", name)
            try:
                self._plot_1d(name)
                pylab.savefig("%s/%s%s.%s"%(self.root_dir, self.prefix, name, self.filetype))
            except Exception as error:
                if self.fatal:
                    raise
                else:
                    print("Unable to plot curve - may be only one value?")
                    print(error)
            finally:
                pylab.close()

    def plot_2d_params(self, names):
        if not names:
            names = self.all_names
        for name1 in names:
            for name2 in names:
                if name1!=name2 and name1<name2:
                    if name1.lower()=='like' or name2.lower()=='like': continue
                    print("Plotting 2D curve for ", name1, "versus", name2)
                    try:
                        self._plot_2d(name1,name2)
                        pylab.savefig("%s/%s%s_%s.%s"%(self.root_dir, self.prefix, name1,name2,self.filetype))
                    except Exception as error:
                        print("(No plot)")
                        print(error)
                    finally:
                        pylab.close()



import sys

"""
This thing is pretty cool - I found it on StackOverflow.

This is a lazily loaded module that provides an interface to pylab.

If you do:

import lazy_pylab as pylab

then you can use pylab in exactly the normal way:
pylab.plot(range(10))
etc., but pylab itself will not be loaded until the first
access of a pylab function or other attribute.

Since pylab can take a relatively long time to load this 
can save time if you never actually end up using it in a
particular run.

"""

class _LazyPylab(object):
    def __init__(self):
        self.first=True
    
    def initialize_matplotlib(self):
        try:
            import logging
            logger = logging.getLogger('matplotlib')
            # set WARNING for Matplotlib
            logger.setLevel(logging.WARNING)
            import matplotlib

        except ImportError:
            print()
            print("No matplotlib: plotting unavailable.")
            print("If you are using postprocess you can disable")
            print("plotting with the --no-plots option")
            print()
            print("Unable to continue cleanly so quitting now.")
            print()
            import sys
            sys.exit(1)
        #Some options are only available in newer
        #matplotlibs.
        try:
            matplotlib.rcParams['figure.max_open_warning'] = 100
        except:
            pass
        matplotlib.rcParams['figure.figsize'] = (8,6)
        matplotlib.rcParams['font.family']='serif'
        matplotlib.rcParams['font.size']=18
        matplotlib.rcParams['legend.fontsize']=15
        matplotlib.rcParams['xtick.major.size'] = 10.0
        matplotlib.rcParams['xtick.minor.size'] = 5.0
        matplotlib.rcParams['ytick.major.size'] = 10.0
        matplotlib.rcParams['ytick.minor.size'] = 5.0
        matplotlib.rcParams['figure.subplot.bottom'] = 0.125
        matplotlib.rcParams['figure.subplot.left'] = 0.175
        matplotlib.use("Agg")

    def __getattr__(self, name):
        if self.first:
            self.initialize_matplotlib()
            self.first=False
        import pylab
        return getattr(pylab, name)


sys.modules[__name__] = _LazyPylab()



from .. import ParallelSampler
import numpy as np
from ...runtime import prior
import scipy.linalg
import types

def abc_model(p):
    #add check that omega_m >0
   for i,pname in enumerate(abc_pipeline.varied_params):
    if pname.name == 'omega_m' and p[i] <0:
        return np.inf

    data = abc_pipeline.run_parameters(p)
    if data is None:
        return None
    #collect the models for each data set we want to use
    model = []
    for like_name in abc_pipeline.likelihood_names:
        #Look in the standard place for the model
        try:
            model.append(data['data_vector', like_name + '_simulation'])
        #Raise an error in the even of failure - this is a systematic problem
        except BlockError:
            raise ValueError("The module in the ABC pipeline should save a model (i.e. a simulation) with the same dimensions as the data, called NAME_simulation.")
    model = np.concatenate(model)
    
    return model


class AbcSampler(ParallelSampler):
    parallel_output = False #True for 
    sampler_outputs = [("weight", float),("distance",float)]

    def config(self):
        try:
            import abcpmc
        except ImportError:
            raise ValueError("To use ABC PMC you need to install it with pip install abcpmc")
        
        global abc_pipeline
        abc_pipeline = self.pipeline

        self.threshold = self.read_ini("threshold",str, 'LinearEps')
        self.metric_kw = self.read_ini("metric",str, 'chi2') #mean, chi2 or other
        if self.metric_kw =='other':
            self.distance_func = self.read_ini("distance_func",str, None) #only for other metric, 
            self.metric = self.distance_func[1:-1]
        self.epimax = self.read_ini('epimax', float,5.0)
        self.epimin = self.read_ini('epimin',float, 1.0)
        self.part_prop = self.read_ini("particle_prop",str,'weighted_cov')
        self.set_prior = self.read_ini("set_prior",str,'uniform')
        self.param_cov = self.read_ini("param_cov_file",str,'None')
        self.knn = self.read_ini("num_nn",int, 10)
        self.npart = self.read_ini("npart",int,100)
        self.niter = self.read_ini("niter",int,2)
        self.ngauss = self.read_ini("ngauss",int,4)
        self.run_multigauss = self.read_ini("run_multigauss",bool,False)
        self.diag_cov = self.read_ini("diag_cov",bool,False)
        self.ndim = len(self.pipeline.varied_params)

        #options for decreasing threshold
        if self.threshold == 'ConstEps':
            self.eps = abcpmc.ConstEps(self.niter, self.epimax)
        elif self.threshold == 'ExpEps':
            self.eps = abcpmc.ExponentialEps(self.niter, self.epimax,self.epimin)
        else:
            self.eps = abcpmc.LinearEps(self.niter, self.epimax, self.epimin)

        print("\nRunning ABC PMC")
        print("with %d particles, %s prior, %s threshold, %d iterations over (%f,%f), %s kernal \n" % (self.npart,self.set_prior,self.threshold,self.niter,self.epimax,self.epimin,self.part_prop))


        #Initial positions for all of the parameters
        self.p0 = np.array([param.start for param in self.pipeline.varied_params])

        #Data file is read for use in dist() for each step
        #parameter covariance used in the prior   
        self.data, self.cov, self.invcov = self.load_data()
        
        #At the moment the same prior (with variable hyperparameters) is
        # used for all parameters  - would be nice to change this to be more flexible 
        self.pmin = np.zeros(self.ndim)
        self.pmax = np.zeros(self.ndim)
        for i,pi in enumerate(self.pipeline.varied_params):
            self.pmin[i] = pi.limits[0]
            self.pmax[i] = pi.limits[1]


        if self.set_prior.lower() == 'uniform':
            self.prior = abcpmc.TophatPrior(self.pmin,self.pmax)
        elif self.set_prior.lower() == 'gaussian':
            sigma2 = np.loadtxt(self.param_cov)
            if len(np.atleast_2d(sigma2)[0][:]) != self.ndim:
                raise ValueError("Cov matrix for Gaussian prior has %d columns for %d params" % len(np.atleast_2d(sigma2)[0][:]), self.ndim)
            else:
                self.prior = abcpmc.GaussianPrior(self.p0, np.atleast_2d(sigma2)) 
        else:
            raise ValueError("Please set the ABC option 'set_prior' to either 'uniform' or 'gaussian'. At the moment only 'uniform' works in the general case.")
        #create sampler
        self.sampler = abcpmc.Sampler(N=self.npart, Y=self.data, postfn=abc_model, dist=self.dist)

        #set particle proposal kernal
        abcpmc.Sampler.particle_proposal_kwargs = {}
        if self.part_prop == 'KNN':
            abcpmc.Sampler.particle_proposal_kwargs = {'k':self.knn}
            self.sampler.particle_proposal_cls = abcpmc.KNNParticleProposal
        elif self.part_prop == 'OLCM':
            self.sampler.particle_proposal_cls = abcpmc.OLCMParticleProposal

        self.converged = False

    def output_samples(self,out):
        for o in out:
            self.output.parameters(o)

    def execute(self):
        # called only by master. Other processors will be in self.pool
        # generate outputs somehow by running log_probability_function
        outputs = []
        for output in self.sampler.sample(self.prior, self.eps):
            outputs.append(np.c_[output.thetas,output.ws,output.dists])
        for out in outputs:
            self.output_samples(out)

        self.converged = True


    def is_converged(self):
        return  self.converged

    def dist(self,x, y):
        if np.isinf(x).any():
            return np.inf
        if self.metric_kw == "chi2":
            d  = x - y
            return np.dot(d,np.dot(self.invcov,d))
        elif self.metric_kw == "mean":
            return np.sum(np.abs(np.mean(x, axis=0) - np.mean(y, axis=0)))
        elif self.metric_kw == "other":
            exec(self.metric)
            return dist_result

    def load_data(self):
        #Load the data by running the pipeline once
        print("Doing and initial ABC run of the pipeline, just to get the data vector from all the likelihood modules.")
        print("This is a slight waste of time (sorry) but the most general way of doing things.")

        block = self.pipeline.run_parameters(self.p0)
        if block is None:
            raise ValueError("I had an error running the pipeline at the central starting values in the values file")

        data = []
        covs = []
        invcovs = []
        for like_name in self.pipeline.likelihood_names:
            data.append(block["data_vector", like_name + "_data"])
            covs.append(block["data_vector", like_name + "_covariance"])
            invcovs.append(block["data_vector", like_name + "_inverse_covariance"])

        data = np.concatenate(data)
        covs = scipy.linalg.block_diag(*covs)
        invcovs = scipy.linalg.block_diag(*invcovs)
        
        if self.diag_cov:
            diag = np.diag(covs)
            covs = np.diag(diag)
            inv = 1./diag
            invcovs = np.diag(inv)

        return  data, covs,invcovs

# legacy alias
ABCSampler = AbcSampler



import subprocess
import os
import tempfile
import shutil
import pytest

this_dir = os.path.split(__file__)[0]
libtest_dir = os.path.join(this_dir, "libtest")
src_dir = os.path.abspath(os.path.join(this_dir, os.pardir))

def test_lib():
    with tempfile.TemporaryDirectory() as tempdir:
        shutil.copytree(libtest_dir, f"{tempdir}/libtest")
        env = os.environ.copy()
        env['COSMOSIS_SRC_DIR'] = src_dir

        # First try making the fortran tester.  If it can't
        # build it at all then we skip the test as our env is
        # presumably not set up to build things properly
        cmd1 = f"cd {tempdir}/libtest && make fortran_t"
        print(cmd1)
        result = subprocess.run(cmd1, shell=True, check=False, env=env)

        if result.returncode:
            pytest.skip("Not in an environment that can compile library tests")

        # If we can compile things, though, then we will want to
        # try compiling everything else and running all the tests
        cmd2 = f"cd {tempdir}/libtest && make test"
        print(cmd2)
        subprocess.run(cmd2, shell=True, check=True, env=env)


if __name__ == '__main__':
    test_lib()



#coding: utf-8
import os
import abc
import sys
import ctypes
from ..datablock import option_section, DataBlock, SectionOptions
from ..utils import underline


MODULE_TYPE_EXECUTE_SIMPLE = "execute"
MODULE_TYPE_EXECUTE_CONFIG = "execute_config"
MODULE_TYPE_SETUP = "setup"
MODULE_TYPE_CLEANUP = "cleanup"

MODULE_LANG_PYTHON = "python"
MODULE_LANG_DYLIB = "dylib"
MODULE_LANG_JULIA = "julia"



class SetupError(Exception):

    u"""Tenuous distinction between generic error and :class:`Module` configuration error.

    :class:`SetupError`, :class:`ValueError` and the generic
    :class:`Exception` should be regarded as synonyms in the context of
    the :class:`Module` class: all indicate some basic problem in the
    .ini files.
"""

    pass



class Module(object):

    u"""Interface to user-defined components of computational pipelines.

    To alleviate confusion, distinguish in your mind the notion of
    :class:`Module` as a cosmosis Module (capital ‘M’ for distinction) as
    opposed to a Python module!

    A :class:`Module` represents a single discrete step in an overall 
    computational pipeline, consuming some inputs from and providing some
    outputs to a :class:`DataBlock`.

    :class:`Module`s are made from either a python file or a shared library,
    and must have a setup function, to be run once when the module is first
    created, and an execute function, to be run whenever a new calculation 
    is to be computed.
    
    This Python class manages the lifetime of a Module, 
    and provides a high-level interface to the module to the rest of the application.

    Unless you are doing something fairly complicated you are unlikely to use
    a module in your own scripts - more generally you would use a :class:`Pipeline`,
    which collects a sequence of modules together.

    The Module /setup/ function, if present, takes in a configuration
    :class:`DataBlock` object (it will find the moduleʼs parameters in a “module_options”
    section), and may return some object which will subsequently be passed
    to the Moduleʼs /execute/ function.

    The /execute/ function itself (which MUST be present in the linked
    library) is also called with a :class:`DataBlock`, and if /setup/
    provided data back to the wrapper, then the /execute/ function MUST
    accept this type of object as its second argument.

    The optional /cleanup/ function is also passed the /setup/ʼs `data` object
    (only), and so may free any resources which that object clings on to.

    """

    def __init__(self, module_name, file_path,
                 setup_function="setup", execute_function="execute",
                 cleanup_function="cleanup", rootpath="."):
        u"""Create an object of type `module_name` from dynamic load library at `file_path`, with interface specified by the `*_function`s.

        The `rootpath` is the directory to search for the linkable
        library.

        Note how `self.*_function`s start out as strings and then become
        executable function objects as the initialization
        progresses—except for `execute_function which is not loaded until
        :func:`setup is called.

        """
        self.name = module_name

        self.setup_function = setup_function
        self.execute_function = execute_function
        self.cleanup_function = cleanup_function

        # identify module filename
        filename = file_path
        if not os.path.isabs(filename):
            filename = os.path.join(rootpath, filename)
        self.filename = filename

        self.library, language = self.load_library(filename)
        self.is_python = (language == MODULE_LANG_PYTHON)
        self.is_julia = (language == MODULE_LANG_JULIA)
        self.is_dynamic = (language == MODULE_LANG_DYLIB)

        # attempt to load setup and cleanup functions
        self.setup_function = self.load_function(self.library,
                                                   setup_function,
                                                   MODULE_TYPE_SETUP,
                                                   set_types=self.is_dynamic,
                                                   )
        self.cleanup_function = self.load_function(self.library,
                                                     cleanup_function,
                                                     MODULE_TYPE_CLEANUP,
                                                     set_types=self.is_dynamic)



    def copy_section_to_module_options(self, config):
        u"""Re-compose the ‘module_options’ section of `config`.

        Remove an existing ‘module_options’ section of `config`, and
        replace it by scanning the entire `config`uration for keys under
        a section named after us and copying their values into a new
        ‘module_options’ section.

        This is done to help the implementation of the module deal with
        its configuration in a generic way, i.e. all modules can simply
        refer to `module_options` for their parameters.

        """
        if config.has_section(option_section):
            config._delete_section(option_section)
        for (section, name) in config.keys(self.name):
            config[option_section, name] = config[section, name]

    def access_check_report(self, config):
        """Check for parameters defined but not used in setup.

        This method scans through the block access logs and records
        all parameters that were accessed.  If any parameters are never
        accessed then it prints a warning.

        The fiddly part of this is accounting for parameters that appear
        in the [DEFAULT] section, since they are usually only used by a
        subset of modules.
        """
        # Read the logs of all parameters that have been read.  This includes
        # everything from all module setups.  We trim it down below.
        nlog = config.get_log_count()
        logs = [config.get_log_entry(i) for i in range(nlog)]

        # get the names of all parameters that came from the
        # DEFAULT section of the config file
        defaults = {k for s,k in config.keys('_cosmosis_default_section')}

        # get all the accesses since the last new-module command
        # The "file" argument doesn't get read during setup because
        # it was used earlier, but should not be in this list.
        # so we explicity include it.
        accesses_by_last_module = {"file"}
        for (log_type, section, name, dtype) in logs:
            # if this is the start of a new module then clear the list
            # because we only want the last one.
            # It's a bit inefficient to loop through the entire
            # list just to get the last chunk of it.  We could go backwards
            # or something like that.  But this is all super fast and only
            # happens once at the start of the pipeline.
            if log_type == "MODULE-START":
                accesses_by_last_module = {"file"}
            # keep only logs that are READs and for the current section
            elif (log_type == "READ-OK") and (section == option_section):
                accesses_by_last_module.add(name)

        # now compare to all the config entries that we have.  If any
        # here are not accessed then there may have been a mistake.
        for (section, name) in config.keys(self.name):
            # Skip the default section.
            if (name not in accesses_by_last_module) and (name not in defaults):
                msg = ("**** WARNING: Parameter '{}' in the [{}] section never used!\n"
                        .format(
                       name, section))
                print(msg)


    def setup_functions(self, config):
        u"""Call the /Module/ constructor.

        This method also pulls in the `execute_function` from the linked
        library.

        This function MUST be run after object initialization and before
        any other action takes place.

        """
        # We need to keep a reference to the full
        # config object for the access check report
        from . import logs
        config_orig = config
        if not self.is_python:
            config = config._ptr

        if self.setup_function:
            logs.overview(underline(f'\nSetting up module {self.name}'))
            self.data = self.setup_function(config)
            self.access_check_report(config_orig)
        else:
            self.data = None

        if self.data is not None:
            module_type = MODULE_TYPE_EXECUTE_CONFIG
        else:
            module_type = MODULE_TYPE_EXECUTE_SIMPLE

        self.execute_function = self.load_function(self.library,
                                                     self.execute_function,
                                                     module_type,
                                                     set_types=self.is_dynamic)
        if self.execute_function is None:
            raise ValueError("Could not find a function 'execute' in module '"
                                 +  self.name + "'")

    def setup(self, config):
        u"""Call the /Module/ after copying config information constructor.
        
        This module is split from the main workhorse method  setup_functions
        so that Modules can be used more easily in external code.
        """
        if isinstance(config, dict):
            config = DataBlock.from_dict(config)
        self.copy_section_to_module_options(config)
        self.setup_functions(config)


    def execute(self, data_block):
        u"""Run the /execute/ function and return whatever it does.

        If the /setup/ function provided some data object, this will be
        passed to the /execute/ function as a second argument.

        """
        if not self.is_python:
            data_block = data_block._ptr
        if not hasattr(self, 'data'):
            raise RuntimeError("Must set up module before executing it")
        if self.data is not None:
            return self.execute_function(data_block, self.data)
        else:
            return self.execute_function(data_block)



    def cleanup(self):
        u"""Run the /cleanup/ function.

        If the /setup/ function provided a data object, this will be
        passed to /cleanup/.

        """
        if self.cleanup_function:
            self.cleanup_function(self.data)



    def __str__(self):
        u"""Return the `name` of this Module."""
        return self.name



    @staticmethod
    def load_library(filepath):
        u"""Whatever kind of file is at `filepath`, try to load it into memory.

        This is really two different functions discriminated by the
        extension on `filepath`: ‘.so’ or ‘.dylib’ will be linked as a
        C-interfaced runtime loadable library, and anything else will be
        taken as a Python module.

        """

        if not os.path.exists(filepath):
            raise SetupError(f"You specified a path {filepath} for a module. "
                             "This file does not exist.")

        elif filepath.endswith('so') or filepath.endswith('dylib'):
            language = MODULE_LANG_DYLIB
            try:
                library = ctypes.cdll.LoadLibrary(filepath)
            except OSError as error:
                raise SetupError(f"You specified a path {filepath} for a module. "
                                 f"File exists, but could not be opened. "
                                 f"Error was {error}")
        elif filepath.endswith(".jl"):
            from . import julia_modules
            library = julia_modules.JuliaModule(filepath)
            language = MODULE_LANG_JULIA
        elif filepath.endswith(".py"):
            language = MODULE_LANG_PYTHON
            dirname, filename = os.path.split(filepath)
            # allows .pyc and .py modules to be used
            impname, ext = os.path.splitext(filename)
            sys.path.insert(0, dirname)
            try:
                library = __import__(impname)
            except ImportError as error:
                raise SetupError("You specified a path %s for a module. "
                                 "I looked for a python module there but "
                                 "was unable to load it.  Error was %s" %
                                 (filepath, error))
            sys.path.pop(0)
        else:
            raise SetupError(f"You specified a path {filepath} for a module. "
                             "I do not know what kind of module this is "
                             "because the suffix is not .so, .dylib, .jl, or .py"
                             )


        return library, language



    @staticmethod
    def load_function(library, function_name,
                      module_type=MODULE_TYPE_EXECUTE_SIMPLE, set_types=True):
        u"""Load a Module's functions from a shared library."""
        function = getattr(library, function_name, None)
        if not function:
            function = getattr(library, function_name + "_", None)

        if function and set_types:
            if module_type == MODULE_TYPE_EXECUTE_SIMPLE:
                function.argtypes = [ctypes.c_voidp]
                function.restype = ctypes.c_int
            elif module_type == MODULE_TYPE_EXECUTE_CONFIG:
                function.argtypes = [ctypes.c_voidp, ctypes.c_voidp]
                function.restype = ctypes.c_int
            elif module_type == MODULE_TYPE_SETUP:
                function.argtypes = [ctypes.c_voidp]
                function.restype = ctypes.c_voidp
            elif module_type == MODULE_TYPE_CLEANUP:
                function.argtypes = [ctypes.c_voidp]
                function.restype = ctypes.c_int
            else:
                raise ValueError("Unknown module type passed to load_interface")
        return function



    @classmethod
    def from_options(cls,module_name,options,root_directory=None):
        u"""The real class constructor.  Return a :class:`Module` based on the contents of user `options`.

        Use the contents of the `module_name` section of the `options` to
        construct a new Module wrapper.

        If `root_directory` (the place to look for Module implementation
        files and maybe associated static data files) is not specified,
        then either the environment variable ‘COSMOSIS_SRC_DIR’ will be
        used, or else the current working directory.

        """
        if root_directory is None:
            root_directory = os.getcwd()

        filename = cls.find_module_file(root_directory,
                                        options.get(module_name, "file"))

        # identify relevant functions
        setup_function = options.get(module_name, "setup", fallback="setup")
        exec_function = options.get(module_name, "function", fallback="execute")
        cleanup_function = options.get(module_name, "cleanup", fallback="cleanup")

        m = cls(module_name, filename,
                setup_function, exec_function, cleanup_function,
                root_directory)

        return m



    @staticmethod
    def find_module_file(base_directory, path):
        u"""Find a module file, which is assumed to be either absolute or relative to COSMOSIS_SRC_DIR."""
        return os.path.join(base_directory, path)





class FunctionModule(Module):
    """
    This subclass, which is designed for when you're using CosmoSIS as a library
    in other applications, lets you manually construct a python module from functions
    you have imported.
    """

    def __init__(self, name, setup_function, execute_function,
                 cleanup_function=None):
        """
        Initialize the subclass from the functions themselves.

        name is a string name.
        
        setup_function, execute_function must be python functions
        matching the expected signature
        def setup(config):
            data = ...
            return data

        def execute(block, config):
            ...
            return 0
    
        """
        self.name = name
        self.filename='missing'

        self.setup_function = setup_function
        self.execute_function = execute_function
        self.cleanup_function = cleanup_function

        self.library = None

        self.is_python = True
        self.is_julia = False
        self.is_dynamic = True

    @staticmethod
    def load_function(library, function_name,
                      module_type=MODULE_TYPE_EXECUTE_SIMPLE, set_types=True):
        return function_name



class ClassModule(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, options):
        pass

    @abc.abstractmethod
    def execute(self, block, config):
        return 0

    def cleanup(config):
        pass

    @classmethod
    def as_module(cls, name):
        def setup(options):
            options = SectionOptions(options)
            mod = cls(options)
            return mod

        def execute(block, mod):
            mod.execute(block)
            return 0

        def cleanup(mod):
            mod.cleanup()


        return FunctionModule(name, setup, execute, cleanup)



from .postprocess import postprocessor_for_sampler
from .inputs import read_input
from .plots import Tweaks
from ..utils import mkdir



def run_cosmosis_postprocess(inputs, **kwargs):
    """
    Run the postprocessing steps on the given input file(s).

    Parameters
    ----------
    inputs : list of str or output objects
        The input file(s) to run postprocessing on.


    """
    #Make the directory for the outputs to go in.
    outdir = kwargs.get("outdir", ".")
    legend = kwargs.get("legend")
    run_max_post = kwargs.get("run_max_post", False)
    text = kwargs.get("text", False)
    weights = kwargs.get("weights", False)
    extra = kwargs.get("extra")
    tweaks = kwargs.get("tweaks")

    mkdir(outdir)
    outputs = {}
    
    #Deal with legends, if any
    if legend:
        labels = legend.split("|")
        if len(labels)!=len(inputs):
            raise ValueError("You specified {} legend names but {} files to plot".format(len(labels), len(inputs)))
    elif isinstance(inputs[0], str):
        labels = inputs
    else:
        labels = [f"chain_{i}" for i in range(len(inputs))]

    if len(inputs)>1 and run_max_post:
        raise ValueError("Can only use the --run-max-post argument with a single parameter file for now")

    processors = []
    processor = None

    for i, ini_filename in enumerate(inputs):
        if "astropy" in str(type(ini_filename)):
            ini_filename.meta.setdefault("chain_name", labels[i])

        sampler, ini = read_input(ini_filename, text, weights)
        processor_class = postprocessor_for_sampler(sampler.split()[-1])

        #We do not know how to postprocess everything.
        if processor_class is None:
            print("I do not know how to postprocess output from the %s sampler"%sampler)
            sampler = None
            continue

        #Create and run the postprocessor

        processor = processor_class(ini, labels[i], i, **kwargs)

        #Inherit any plots from the previous postprocessor
        #so we can make plots with multiple datasets on
        processor.outputs.update(outputs)

        #We can load extra plots to make from a python
        #script here
        if extra:
            processor.load_extra_steps(extra)

        #Optionally add a step in which we re-run the best-fit
        if run_max_post:
            processor.add_rerun_bestfit_step(run_max_post)

        #Run the postprocessor and make the outputs for this chain
        processor.run()

        #Save the outputs ready for the next post-processor in case
        #they want to add to it (e.g. two constriants on the same axes)
        outputs.update(processor.outputs)

    # If there was no successul postprocessor then we are done
    if processor is None:
        return

    # Finalize all the elements - this adds legends to any plots
    # that need them. This final processor knows about all the
    # outputs that we made.
    processor.finalize()

    if sampler is None:
        return

    #Run any tweaks that the user specified
    if tweaks:
        tweaks = Tweaks.instances_from_file(tweaks)
        for tweak in tweaks:
            processor.apply_tweaks(tweak)

    return processor



import multiprocessing
import os


class Pool(object):
    def __init__(self, processes):
        self.size = processes
        self.rank = 0
        self.master_pid = os.getpid()

    def is_master(self):
        return self.master_pid == os.getpid()

    def map(self, function, args):
        with multiprocessing.Pool(self.size) as pool:
            results = pool.map(function, args)
        return results

    def close(self):
        pass

    def bcast(self, data):
        return self.data

    def gather(self, data):
        return self.data

    def allreduce(self, data):
        return self.data

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass        


#!/usr/bin/env python
import matplotlib
matplotlib.use("Agg")
import pylab
import numpy as np


data = np.loadtxt("demo3_output.txt").T

r_T = data[0]
log_like = data[1]

#convert to like from log-like and normalize
like = np.exp(log_like)
like /= like.max()

#make plot
pylab.plot(r_T, like)
pylab.xlabel("Tensor ratio r")
pylab.ylabel("Likelihood")
pylab.grid()

#Save
pylab.savefig("plots/demo3.png")



from .. import ParallelSampler
from ...runtime import logs
import numpy as np
import os


def log_probability_function(p):
    r = pipeline.run_results(p)
    out = [r.like, r.prior]

    # Flatten any vector outputs here
    for e in r.extra:
        if np.isscalar(e):
            out.append(e)
        else:
            out.extend(e)
    out = tuple(out)
    return out


def prior_transform(p):
    return pipeline.denormalize_vector_from_prior(p)


class NautilusSampler(ParallelSampler):
    parallel_output = False
    internal_resume = True
    sampler_outputs = [('log_weight', float), ('prior', float),
                       ("post", float)]

    def config(self):
        global pipeline
        pipeline = self.pipeline

        if self.is_master():
            self.n_live = self.read_ini("n_live", int, 2000)
            self.n_update = self.read_ini("n_update", int, self.n_live)
            self.enlarge_per_dim = self.read_ini("enlarge_per_dim", float, 1.1)
            self.n_points_min = self.read_ini("n_points_min", int,
                                              self.pipeline.nvaried + 50)
            self.split_threshold = self.read_ini("split_threshold", float,
                                                 100.0)
            self.n_networks = self.read_ini("n_networks", int, 4)
            self.n_batch = self.read_ini("n_batch", int, 100)
            self.seed = self.read_ini("seed", int, -1)
            if self.seed < 0:
                self.seed = None
            self.resume_ = self.read_ini("resume", bool, False)
            self.f_live = self.read_ini("f_live", float, 0.01)
            self.n_shell = self.read_ini("n_shell", int, self.n_batch)
            self.n_eff = self.read_ini("n_eff", float, 10000.0)
            self.n_like_max = self.read_ini("n_like_max", int, -1)
            if self.n_like_max < 0:
                self.n_like_max = np.inf
            self.discard_exploration = self.read_ini(
                "discard_exploration", bool, False)
            self.verbose = self.read_ini("verbose", bool, False)

        self.converged = False

    def execute(self):
        from nautilus import Sampler

        n_dim = self.pipeline.nvaried

        try:
            resume_filepath = self.output.name_for_sampler_resume_info()
        except NotImplementedError:
            resume_filepath = None

        if resume_filepath is not None:
            resume_filepath = resume_filepath + ".hdf5"
            if self.resume_ and os.path.exists(resume_filepath):
                if self.is_master():
                    logs.overview(f"Resuming Nautilus from file {resume_filepath}")

        sampler = Sampler(
            prior_transform,
            log_probability_function,
            n_dim,
            n_live=self.n_live,
            n_update=self.n_update,
            enlarge_per_dim=self.enlarge_per_dim,
            n_points_min=self.n_points_min,
            split_threshold=self.split_threshold,
            n_networks=self.n_networks,
            n_batch=self.n_batch,
            seed=self.seed,
            filepath=resume_filepath,
            resume=self.resume_,
            pool=self.pool,
            blobs_dtype=float
        )

        sampler.run(f_live=self.f_live,
                    n_shell=self.n_shell,
                    n_eff=self.n_eff,
                    n_like_max=self.n_like_max,
                    discard_exploration=self.discard_exploration,
                    verbose=self.verbose)

        results = sampler.posterior(return_blobs=True)
        if isinstance(results[3][0], float):
            priors = results[3]
        else:
            priors = np.array([r[0] for r in results[3]])

        posts = results[2] + priors
        self.distribution_hints.set_from_sample(results[0], posts, log_weights=results[1])

        for sample, logwt, logl, blob in zip(*results):
            if isinstance(blob, float):
                prior = blob
                self.output.parameters(sample, logwt, prior, logl + prior)
            else:
                prior = blob[0]
                extra = list(blob)[1:]
                self.output.parameters(sample, extra, logwt, prior, logl + prior)

        self.output.final("efficiency", sampler.n_eff / sampler.n_like)
        self.output.final("neff", sampler.n_eff)
        self.output.final("nsample", len(sampler.posterior()[0]))
        self.output.final("log_z", sampler.log_z)
        self.output.final("log_z_error", 1.0 / np.sqrt(sampler.n_eff))
        self.converged = True

    def is_converged(self):
        return self.converged




from cosmosis.datablock import option_section
# from cosmosis.datablock.cosmosis_py.block import option_section
    
def mksetup(cls):
    def setup(config):
        return cls(config,option_section)
    return setup

def execute(block, state):
    return state.execute(block)

def cleanup(state):
    return state.cleanup()
    
def declare_module(class_obj):
    tmp = __import__(class_obj.__dict__['__module__'])
    tmp.__dict__['execute']=execute
    tmp.__dict__['setup']=mksetup(class_obj)
    tmp.__dict__['cleanup']=cleanup
    tmp.__dict__['pymodule']=class_obj



import collections
import types


enum_names = [
"DBS_SUCCESS",
"DBS_DATABLOCK_NULL",
"DBS_SECTION_NULL",
"DBS_SECTION_NOT_FOUND",
"DBS_NAME_NULL",
"DBS_NAME_NOT_FOUND",
"DBS_NAME_ALREADY_EXISTS",
"DBS_VALUE_NULL",
"DBS_WRONG_VALUE_TYPE",
"DBS_MEMORY_ALLOC_FAILURE",
"DBS_SIZE_NULL",
"DBS_SIZE_NONPOSITIVE",
"DBS_SIZE_INSUFFICIENT",
"DBS_NDIM_NONPOSITIVE",
"DBS_NDIM_OVERFLOW",
"DBS_NDIM_MISMATCH",
"DBS_EXTENTS_NULL",
"DBS_EXTENTS_MISMATCH",
"DBS_LOGIC_ERROR"
]


class EnumGenerator(object):
    """This is a slightly insane hack to generate an enum into the local
    namespace from a list of names.  This avoids a nested namespace.
    Once the list of errors has settled down so we do not need to keep
    retyping we can just hard-code them.
    """
    def __init__(self, namespace):
        self.namespace=namespace
        self.value = 0
    def add(self, name):
        self.namespace[name] = self.value
        self.value += 1

enum = EnumGenerator(locals())
for name in enum_names:
    enum.add(name)

ERROR_MESSAGES = collections.defaultdict(lambda:  "{status}: Uknown error; possibly cosmosis internal failure.  Please contact cosmosis team")
ERROR_MESSAGES.update({
    DBS_DATABLOCK_NULL: "{status}: Null (zero) datablock passed into function (section was {section}, name was {name})",
    DBS_SECTION_NULL: "{status}: Null name passed into function (section was {section}",
    DBS_SECTION_NOT_FOUND: "{status}: Could not find section called {section} (name was {name})",
    DBS_NAME_NULL: "{status}: Null name passed into function (section was {section})",
    DBS_NAME_NOT_FOUND: "{status}: Could not find name {name} in section {section}",
    DBS_NAME_ALREADY_EXISTS: "{status}: Tried to overwrite {name} in section {section}.  Use the replace functions to over-write",
    DBS_VALUE_NULL: "{status}: Passed a null value into function (section was {section}, name was {name})",
    DBS_WRONG_VALUE_TYPE: "{status}: Wrong value type for {name} in section {section}",
    DBS_MEMORY_ALLOC_FAILURE: "{status}: Failed to allocate memory for {name} in section {section}.",
    DBS_SIZE_NULL: "{status}: Null parameter for returned size passed into function (section was {section}, name was {name})",
    DBS_SIZE_NONPOSITIVE: "{status}: Non-positive maximum size passed into function (section was {section}, name was {name})",
    DBS_SIZE_INSUFFICIENT: "{status}: Size of passed in array not large enough for what is needed (section was {section}, name was {name})",
    DBS_NDIM_NONPOSITIVE: "{status}: Number of dimensions not positive (section was {section}, name was {name})",
    DBS_NDIM_OVERFLOW: "{status}: Number of dimensions in array is larger than can be represented in an 'int' (section was {section}, name was {name})",
    DBS_NDIM_MISMATCH: "{status}: Number of dimensions supplied for array does not match the array dimensions (section was {section}, name was {name})",
    DBS_EXTENTS_NULL: "{status} Null value passed for array extents (section was {section}, name was {name})",
    DBS_EXTENTS_MISMATCH: "{status} Supplied array extents do not match the extents of the stored array (section was {section}, name was {name})",
    DBS_LOGIC_ERROR: "{status}: Internal cosmosis logical error.  Please contact cosmosis team (section was {section}, name was {name})",
})

ERROR_CLASSES = {}

class CosmosisError(Exception):
    pass

 #Might want to create more of a subclass structure here.
 #Have them all inherit from BlockError
class BlockError(CosmosisError):
    def __init__(self, status, section, name):
        self.name=name
        self.status=status
        self.section=section

    @staticmethod
    def exception_for_status(status, section, name):
        return ERROR_CLASSES[status](status, section, name)
    
    def __str__(self):
        return ERROR_MESSAGES[self.status].format(status=self.status, name=self.name,section=self.section)


def underscore_to_camelcase(value):
    return "".join(str.capitalize(x) if x else '_' for x in value.split("_"))

#for each status except the success one we create an error class
for name in enum_names[1:]:
    class_name = 'Block'+underscore_to_camelcase(name[4:].lower())
    cls = type(class_name, (BlockError,), {})
    status = locals()[name]
    locals()[class_name] = cls
    ERROR_CLASSES[status] = cls 




from ...runtime import logs
from .. import ParallelSampler
import numpy as np
import sys


def log_probability_function(p):
    results = pipeline.run_results(p)
    # also save the prior because we want the posterior which dynesty does not report
    blob = np.concatenate([[results.prior], results.extra])
    return results.like, blob

def prior_transform(p):
    return pipeline.denormalize_vector_from_prior(p)

class DynestySampler(ParallelSampler):
    parallel_output = False
    sampler_outputs = [('log_weight', float), ("prior", float), ("post", float)]

    def config(self):
        global pipeline
        pipeline = self.pipeline

        if self.is_master():
            self.mode = self.read_ini_choices("mode", str, ["static", "dynamic"], "static")
            self.nlive = self.read_ini("nlive", int, 500)
            self.bound = self.read_ini_choices("bound", str, ["none", "single", "multi", "balls", "cube"], "multi")
            self.sample = self.read_ini_choices("sample", str, ["unif", "rwalk", "slice", "rslice", "hslice", "auto"], "auto")
            self.update_interval = self.read_ini("update_interval", float, 0.6)
            self.min_ncall = self.read_ini("min_ncall",int, 2*self.nlive)
            self.min_eff = self.read_ini("min_eff",float, 10.0)
            default_queue_size = 1 if self.pool is None else self.pool.size
            self.queue_size = self.read_ini("queue_size", int, default_queue_size)
            self.parallel_prior = self.read_ini("parallel_prior", bool, True)
            self.max_call = self.read_ini("max_call", int, sys.maxsize)
            self.dlogz = self.read_ini("dlogz", float, 0.01)
            print_progress_default = logs.is_enabled_for(logs.logging.INFO)
            self.print_progress = self.read_ini("print_progress", bool, print_progress_default)

        self.converged = False



    def execute(self):
        from dynesty import NestedSampler, DynamicNestedSampler

        ndim = self.pipeline.nvaried


        if self.mode == "static":
            sampler = NestedSampler(
                log_probability_function,
                prior_transform, 
                ndim, 
                nlive = self.nlive,
                bound = self.bound,
                sample = self.sample,
                update_interval = self.update_interval,
                first_update = {'min_ncall':self.min_ncall, 'min_eff':self.min_eff},
                queue_size = self.queue_size,
                pool = self.pool,
                blob=True
                )

            sampler.run_nested(dlogz=self.dlogz)

        else:
            sampler = DynamicNestedSampler(
                log_probability_function,
                prior_transform, 
                ndim, 
                bound = self.bound,
                sample = self.sample,
                # update_interval = self.update_interval,
                queue_size = self.queue_size,
                pool = self.pool,
                blob=True
                )
            sampler.run_nested(dlogz_init=self.dlogz)

        results = sampler.results
        results.summary()

        posts = results['blob'][:, 0] + results['logl']
        self.distribution_hints.set_from_sample(results['samples'], posts, log_weights=results['logwt'])

        for sample, logwt, logl, derived in zip(results['samples'],results['logwt'], results['logl'], results['blob']):
            prior = derived[0]
            post = prior + logl
            self.output.parameters(sample, derived[1:], logwt, prior, post)

        self.output.final("efficiency", results['eff'])
        self.output.final("nsample", len(results['samples']))
        self.output.final("log_z", results['logz'][-1])
        self.output.final("log_z_error", results['logzerr'][-1])
        self.converged = True



    def is_converged(self):
        return self.converged
